# coding=utf-8
from __future__ import division

from . import ph, TestData
from math import ceil as _ceil
import numpy as np

import unittest

__author__ = 'aleb'


# noinspection PyArgumentEqualDefault
class GeneralTest(unittest.TestCase):
    def test_fmap_and_algo(self):
        s = ph.EvenlySignal((np.sin(np.random.rand(100) * 3.14 - (3.14 / 2)) + 1) * 93, 15)
        lf = ph.PowerInBand(freq_max=1, freq_min=0.001, method='ar')
        hf = ph.PowerInBand(freq_max=4, freq_min=1, method='ar')

        algos = [
            ph.Mean(),
            ph.StDev(),
            ph.NNx(threshold=100),
            ph.PowerInBand(interp_freq=20, freq_max=4, freq_min=0.001, method='ar'),
            ph.PowerInBand(interp_freq=20, freq_min=4, freq_max=15, method='ar'),
            ph.algo(lambda d, p: lf(d) / hf(d))(),
            ph.algo(lambda d, p: len(d))()
        ]

        g = ph.FixedSegments(step=1, width=1.5)
        l = len([x for x in g(s)])
        r, ignored = ph.fmap(g(s), algos)
        n = 3 + len(algos)
        for i in r:
            self.assertEquals(len(i), n)

        self.assertEquals(len(r), l)

    def test_fmap(self):
        samples = 1000
        freq_down = 13
        freq = freq_down * 7
        start = 1460713373
        nature = "una_bif-fa"
        np.random.seed(1234)

        s = ph.EvenlySignal(values=np.cumsum(np.random.rand(1, samples) - .5) * 100,
                            sampling_freq=freq,
                            signal_nature=nature,
                            start_time=start,
                            )

        features = [
            ph.PNNx(threshold=50),
            ph.PNNx(threshold=25),
            ph.PNNx(threshold=10),
            ph.PNNx(threshold=5),
            ph.Mean(),
            ph.StDev(),
            ph.Median(),
        ]

        segmenter = ph.FixedSegments(width=3, step=2)
        segments = [i for i in segmenter(s)]

        results, columns = ph.fmap(segments, features, s)

        self.assertEqual(results.shape, (len(segments), len(features) + 3))
        self.assertEqual(len(columns), len(features) + 3)

    def test_ex_more(self):
        s = ph.EvenlySignal(np.cumsum(np.random.rand(1000) - .5) * 100, 10)

        w1 = ph.FixedSegments(step=2, width=3)(s)
        w3 = ph.LabelSegments(labels=ph.UnevenlySignal(
            values=['a', 'a', 'b', 'a', 'r', 's', 'r', 'b'],
            x_values=[10, 12, 13.5, 14.3, 15.6, 20.1123, 25, 36.8],
            sampling_freq=10,
            start_time=8,
            x_type='instants'))(s)
        w3i = [x for x in w3]
        w4 = ph.CustomSegments(begins=[x.get_begin_time() for x in w3i],
                               ends=[x.get_end_time() for x in w3i])(s)
        w4i = [x for x in w4]

        self.assertEqual(len(w4i), len(w3i))

        y1 = [x for x in w1]
        y3 = [x for x in w3]
        ph.fmap(w1, [ph.Mean(), ph.StDev(), ph.NNx(threshold=31)])
        ph.fmap(w3, [ph.Mean(), ph.StDev(), ph.NNx(threshold=31)])
        ph.fmap(y1, [ph.Mean(), ph.StDev(), ph.NNx(threshold=31)])
        ph.fmap(y3, [ph.Mean(), ph.StDev(), ph.NNx(threshold=31)])

        # noinspection PyArgumentEqualDefault
        sd2 = s.resample(1, 'linear')
        sd3 = s.resample(1, 'nearest')
        sd4 = s.resample(1, 'zero')
        sd5 = s.resample(1, 'slinear')
        sd6 = s.resample(1, 'quadratic')
        sd7 = s.resample(1, 'cubic')

        self.assertEqual(len(sd2), 100)
        self.assertEqual(len(sd3), 100)
        self.assertEqual(len(sd4), 100)
        self.assertEqual(len(sd5), 100)
        self.assertEqual(len(sd6), 100)
        self.assertEqual(len(sd7), 100)

        # noinspection PyArgumentEqualDefault
        so2 = s.resample(20, 'linear')
        so3 = s.resample(20, 'nearest')
        so4 = s.resample(20, 'zero')
        so5 = s.resample(20, 'slinear')
        so6 = s.resample(20, 'quadratic')
        so7 = s.resample(20, 'cubic')

        self.assertEqual(len(so2), 2000)
        self.assertEqual(len(so3), 2000)
        self.assertEqual(len(so4), 2000)
        self.assertEqual(len(so5), 2000)
        self.assertEqual(len(so6), 2000)
        self.assertEqual(len(so7), 2000)

        so1 = s.resample(21)

        self.assertEqual(len(so1), 2100)

    def test_evenly_signal_base(self):
        samples = 1000
        freq_down = 13
        freq = freq_down * 7
        start = 1460713373
        nature = "una_bif-fa"

        s = ph.EvenlySignal(values=np.cumsum(np.random.rand(1, samples) - .5) * 100,
                            sampling_freq=freq,
                            signal_nature=nature,
                            start_time=start,
                            )

        assert s.get_sampling_freq() == freq
        assert s.get_signal_nature() == nature
        assert s.get_start_time() == start

        # ineffective
        s.resample(freq_down)

        # FIXME: this is part of test_evenly_signal_resample (?)
        # assert properties of resampled s

        # length Y
        self.assertEqual(len(s), samples)
        self.assertEqual(len(s.get_values()), samples)
        # length X
        self.assertEqual(len(s.get_times()), samples)
        # sampling frequency in Hz
        self.assertEqual(s.get_sampling_freq(), freq)
        # duration in seconds
        self.assertAlmostEqual(s.get_duration(), samples / freq, 6)
        # start timestamp
        self.assertEqual(s.get_start_time(), start)
        # end timestamp
        self.assertEqual(s.get_end_time(), start + s.get_duration())
        # start time
        self.assertEqual(s.get_signal_nature(), nature)

    # TODO : test_unevenly_signal_base with x_values of type 'indices'
    def test_unevenly_signal_base(self):
        samples = 200
        freq = 13
        start = 0
        nature = "una_bif-fa"

        s = ph.UnevenlySignal(values=np.cumsum(np.random.rand(1, samples) - .5) * 100,
                              x_values=np.cumsum(np.random.rand(1, samples)) * 100 + start,
                              sampling_freq=freq,
                              signal_nature=nature,
                              start_time=start,
                              x_type='indices'
                              )
        # assert properties of original s
        #
        # length Y
        self.assertEqual(len(s), samples)
        self.assertEqual(len(s.get_values()), samples)
        # length X
        self.assertEqual(len(s.get_indices()), samples)
        # sampling frequency in Hz
        self.assertEqual(s.get_sampling_freq(), freq)
        # start timestamp
        self.assertEqual(s.get_start_time(), start)
        # end timestamp
        self.assertEqual(s.get_end_time(), start + s.get_duration())
        # start time
        self.assertEqual(s.get_signal_nature(), nature)

    def test_evenly_signal_resample(self):
        samples = 1000
        freq_down = 7
        freq = freq_down * 13
        freq_up = freq * 3
        freq_down_r = freq / 5
        start = 1460713373
        nature = "una_bif-fa"

        s = ph.EvenlySignal(values=np.cumsum(np.random.rand(1, samples) - .5) * 100,
                            sampling_freq=freq,
                            signal_nature=nature,
                            start_time=start,
                            )

        # TODO: test also different interpolationi methods
        def check_resampled(freq_new):
            # resample
            resampled = s.resample(freq_new)
            # new number of samples
            new_samples = _ceil(samples * freq_new / freq)

            # length Y
            self.assertEqual(len(resampled), new_samples)
            # length X
            self.assertEqual(len(resampled.get_times()), len(resampled))
            # sampling frequency in Hz
            self.assertEqual(resampled.get_sampling_freq(), freq_new)
            # duration differs less than 1s from the original
            self.assertLess(abs(resampled.get_duration() - samples / freq), 1)
            # start timestamp
            self.assertEqual(resampled.get_start_time(), start)
            # end timestamp
            self.assertEqual(resampled.get_end_time(), start + resampled.get_duration())
            # start time
            self.assertEqual(resampled.get_signal_nature(), nature)

        # down-sampling int
        check_resampled(freq_down)
        # over-sampling int
        check_resampled(freq_up)
        # down-sampling rationale
        check_resampled(freq_down_r)

    def test_unevenly_signal_to_evenly(self):
        samples = 200
        indexes = np.cumsum(np.round(np.random.rand(1, samples) + 1))
        freq = 13
        start = 0
        nature = "una_bif-fa"

        s = ph.UnevenlySignal(values=np.cumsum(np.random.rand(1, samples) - .5) * 100,
                              x_values=indexes,
                              sampling_freq=freq,
                              signal_nature=nature,
                              start_time=start,
                              x_type='indices'
                              )

        # conversion
        s = s.to_evenly()

        # length
        self.assertEqual(len(s.get_values()), len(s))
        # sampling frequency in Hz
        self.assertEqual(s.get_sampling_freq(), freq)
        # start timestamp
        self.assertEqual(s.get_start_time(), s.get_time_from_iidx(0))
        # end timestamp
        self.assertEqual(s.get_end_time(), s.get_time_from_iidx(0) + s.get_duration())
        # start time
        self.assertEqual(s.get_signal_nature(), nature)

    def test_segmentation(self):
        samples = 1000
        freq_down = 7
        freq = freq_down * 13
        start = 1460000000
        nature = "una_bif-fa"

        s = ph.EvenlySignal(values=np.cumsum(np.random.rand(1, samples) - .5) * 100,
                            sampling_freq=freq,
                            signal_nature=nature,
                            start_time=start
                            )

        def function(y):
            for j in range(len(y)):
                self.assertLessEqual(y[j].get_begin(), y[j].get_end())
                self.assertGreaterEqual(y[j].get_begin_time(), s.get_start_time())
                self.assertGreaterEqual(y[j].get_end_time(), s.get_start_time())
                # # list[x:y] where y > len(list) is a valid usage in python so next lines are cut
                # self.assertLessEqual(y[j].get_begin(), len(s))
                # self.assertLessEqual(y[j].get_end(), len(s))

        w1 = ph.FixedSegments(step=2, width=3)(s)
        y1 = [x for x in w1]
        function(y1)

        w2 = ph.FixedSegments(step=100, width=121)(s)
        y2 = [x for x in w2]
        function(y2)

        w3 = ph.LabelSegments(labels=ph.UnevenlySignal(values=[0, 0, 1, 0, 2, 3, 2, 1],
                                                       x_values=np.array(
                                                           [10, 12, 13.5, 14.3, 15.6, 20.1123, 25, 36.8]) + start,
                                                       # start_time=start,
                                                       x_type='instants'
                                                       ))(s)

        y3 = [x for x in w3]
        function(y3)

        w4 = ph.CustomSegments(begins=np.arange(s.get_start_time(), s.get_end_time(), 3),
                               ends=np.arange(s.get_start_time() + 1, s.get_end_time(), 3))(s)
        y4 = [x for x in w4]
        function(y4)

        r = [ph.fmap(w1, [ph.Mean(), ph.StDev(), ph.NNx(threshold=31)]),
             ph.fmap(w2, [ph.Mean(), ph.StDev(), ph.NNx(threshold=31)]),
             ph.fmap(w3, [ph.Mean(), ph.StDev(), ph.NNx(threshold=31)]),
             ph.fmap(w4, [ph.Mean(), ph.StDev(), ph.NNx(threshold=31)]),
             ph.fmap(y1, [ph.Mean(), ph.StDev(), ph.NNx(threshold=31)]),
             ph.fmap(y2, [ph.Mean(), ph.StDev(), ph.NNx(threshold=31)]),
             ph.fmap(y3, [ph.Mean(), ph.StDev(), ph.NNx(threshold=31)]),
             ph.fmap(y4, [ph.Mean(), ph.StDev(), ph.NNx(threshold=31)])]

        n = 4
        for i in range(n):
            self.assertEqual(len(r[i]), len(r[i + n]))

    def test_cache_with_time_domain(self):
        samples = 1000
        freq_down = 13
        freq = freq_down * 7
        start = 1460713373
        nature = "una_bif-fa"

        s = ph.EvenlySignal(values=np.cumsum(np.random.rand(1, samples) - .5) * 100,
                            sampling_freq=freq,
                            signal_nature=nature,
                            start_time=start,
                            )

        def function():
            # Mean
            self.assertEqual(np.nanmean(s), ph.Mean()(s))
            # Min
            self.assertEqual(np.nanmin(s), ph.Min()(s))
            # Max
            self.assertEqual(np.nanmax(s), ph.Max()(s))
            # Range
            self.assertEqual(np.nanmax(s) - np.nanmin(s), ph.Range()(s))
            # Median
            self.assertEqual(np.median(s), ph.Median()(s))
            # StDev
            self.assertEqual(np.nanstd(s), ph.StDev()(s))

        # raw
        function()
        # cache
        function()

    def test_evenly_slicing(self):
        samples = 1000
        x1 = 200
        x2 = 201
        x3 = 590
        x4 = 999
        freq_down = 7
        freq = freq_down * 13
        start = 1460713373
        nature = "una_bif-fa"

        s = ph.EvenlySignal(values=np.cumsum(np.random.rand(1, samples) - .5) * 100,
                            sampling_freq=freq,
                            signal_nature=nature,
                            start_time=start,
                            )

        s1 = s.segment_idx(0, x4)
        s2 = s.segment_idx(x1, x3)
        s3 = s.segment_idx(x2, -1)
        s4 = s.segment_idx(x3, samples)
        s5 = s.segment_idx(0, None)
        s6 = s.segment_idx(None, x2)
        s7 = s.segment_idx(None, None)

        self.assertEqual(len(s1), x4 - 0)
        self.assertEqual(len(s2), x3 - x1)
        self.assertEqual(len(s3), samples - x2 - 1)
        self.assertEqual(len(s4), samples - x3)
        self.assertEqual(len(s5), samples)
        self.assertEqual(len(s6), x2)
        self.assertEqual(len(s7), samples)
        self.assertEqual(s1.get_start_time(), s.get_time(0))
        self.assertEqual(s2.get_start_time(), s.get_time(x1))
        self.assertEqual(s3.get_start_time(), s.get_time(x2))
        self.assertEqual(s4.get_start_time(), s.get_time(x3))
        self.assertEqual(s5.get_start_time(), s.get_time(0))
        self.assertEqual(s6.get_start_time(), s.get_time(0))
        self.assertEqual(s7.get_start_time(), s.get_time(0))
        self.assertEqual(s1.get_end_time(), s1.get_time(x4 - 0 - 1) + 1. / s1.get_sampling_freq())
        self.assertEqual(s2.get_end_time(), s2.get_time(x3 - x1 - 1) + 1. / s2.get_sampling_freq())
        self.assertEqual(s3.get_end_time(), s3.get_time(samples - x2 - 1 - 1) + 1. / s3.get_sampling_freq())
        self.assertEqual(s4.get_end_time(), s4.get_time(samples - x3 - 1) + 1. / s4.get_sampling_freq())
        self.assertEqual(s5.get_end_time(), s5.get_time(samples - 1) + 1. / s5.get_sampling_freq())
        self.assertEqual(s6.get_end_time(), s6.get_time(x2 - 1) + 1. / s6.get_sampling_freq())
        self.assertEqual(s7.get_end_time(), s7.get_time(samples - 1) + 1. / s7.get_sampling_freq())

    def test_unevenly_slicing(self):
        samples = 1000
        x1 = 200
        x2 = 201
        x3 = 590
        x4 = 999
        freq = 7 * 13
        start = 0
        nature = "una_bif-fa"

        x_vals = np.cumsum(np.random.rand(1, samples) * 9 + 1).astype(int)

        s = ph.UnevenlySignal(values=np.cumsum(np.random.rand(1, samples) - .5) * 100,
                              x_values=x_vals,
                              sampling_freq=freq,
                              signal_nature=nature,
                              start_time=start,
                              x_type='indices'
                              )

        s1 = s.segment_iidx(0, x4)
        s2 = s.segment_iidx(x1, x3)
        s3 = s.segment_iidx(x2, -1)
        s4 = s.segment_iidx(x3, samples)
        s5 = s.segment_iidx(0, None)
        s6 = s.segment_iidx(None, x2)
        s7 = s.segment_iidx(None, None)

        self.assertEqual(len(s1), x4 - 0)
        self.assertEqual(len(s2), x3 - x1)
        self.assertEqual(len(s3), samples - x2 - 1)
        self.assertEqual(len(s4), samples - x3)
        self.assertEqual(len(s5), samples)
        self.assertEqual(len(s6), x2)
        self.assertEqual(len(s7), samples)

        self.assertEqual(len(s1), len(s1.get_indices()))
        self.assertEqual(len(s2), len(s2.get_indices()))
        self.assertEqual(len(s3), len(s3.get_indices()))
        self.assertEqual(len(s4), len(s4.get_indices()))
        self.assertEqual(len(s5), len(s5.get_indices()))
        self.assertEqual(len(s6), len(s6.get_indices()))
        self.assertEqual(len(s7), len(s7.get_indices()))

        self.assertEqual(s1.get_indices()[0], 0)
        self.assertEqual(s2.get_indices()[0], 0)
        self.assertEqual(s3.get_indices()[0], 0)
        self.assertEqual(s4.get_indices()[0], 0)
        self.assertEqual(s5.get_indices()[0], 0)
        self.assertEqual(s6.get_indices()[0], 0)
        self.assertEqual(s7.get_indices()[0], 0)
        self.assertEqual(s1.get_indices()[-1], s.get_indices()[x4 - 1] - s.get_indices()[0])
        self.assertEqual(s2.get_indices()[-1], s.get_indices()[x3 - 1] - s.get_indices()[x1])
        self.assertEqual(s3.get_indices()[-1], s.get_indices()[-1 - 1] - s.get_indices()[x2])
        self.assertEqual(s4.get_indices()[-1], s.get_indices()[samples - 1] - s.get_indices()[x3])
        self.assertEqual(s5.get_indices()[-1], s.get_indices()[samples - 1] - s.get_indices()[0])
        self.assertEqual(s6.get_indices()[-1], s.get_indices()[x2 - 1] - s.get_indices()[0])
        self.assertEqual(s7.get_indices()[-1], s.get_indices()[samples - 1] - s.get_indices()[0])

    # noinspection PyMethodMayBeStatic
    def test_signal_plot(self):
        e = ph.EvenlySignal(values=TestData.ecg()[:10000], sampling_freq=1024, signal_nature="ecg")
        e, ignored, ignored, ignored = ph.PeakDetection(delta=1)(e)
        e = ph.UnevenlySignal(values=e, x_values=e, x_type='indices', sampling_freq=1024, signal_nature="ibi")
        e.plot("|b")
        e = ph.EvenlySignal(values=TestData.eda()[:10000], sampling_freq=1024, signal_nature="gsr")
        e.plot()
        e = ph.EvenlySignal(values=TestData.bvp()[:10000], sampling_freq=1024, signal_nature="bvp")
        e.plot()
        e = ph.EvenlySignal(values=TestData.resp()[:10000], sampling_freq=1024, signal_nature="resp")
        e.plot()

        # import matplotlib.pyplot as plt
        # plt.show()
        # plt.close('all')


if __name__ == '__main__':
    unittest.main()



# coding=utf-8
from __future__ import division
import numpy as _np
from scipy import interpolate as _interp
from matplotlib.pyplot import plot as _plot, vlines as _vlines, xlabel as _xlabel, ylabel as _ylabel, grid as _grid 
from matplotlib.pyplot import subplot as _subplot, tight_layout as _tight_layout, subplots_adjust as _subplots_adjust, xlim as _xlim
from numbers import Number as _Number
from pyphysio.Utility import abstractmethod as _abstract, PhUI as _PhUI
import copy
#from pyphysio.filters.Filters import ImputeNAN as _ImputeNAN
__author__ = 'AleB'


# TODO: Consider collapsing classes

def from_pickleable(pickle):
    """
    Builds a Signal using the pickleable tuple version of it.
    :param pickle: Tuple of the form (Signal, ph dict).
    :return: Signal
    """
    d, ph = pickle
    assert isinstance(d, Signal)
    assert isinstance(ph, dict)
    d._pyphysio = ph
    return d


def from_pickle(path):
    """
    Loads a Signal from a pickle file given the path.
    :param path: File system path to the pickle file.
    :return: A Signal.
    """
    from gzip import open
    from pickle import load
    f = open(path)
    p = load(f)
    f.close()
    return from_pickleable(p)


class Signal(_np.ndarray):
    _MT_NATURE = "signal_type"
    _MT_START_TIME = "start_time"
    _MT_SAMPLING_FREQ = "sampling_freq"
    _MT_INFO_ATTR = "_pyphysio"

    def __new__(cls, values, sampling_freq, start_time=None, signal_type=""):
        assert sampling_freq > 0, "The sampling frequency cannot be zero or negative"
        assert start_time is None or isinstance(start_time, _Number), "Start time is not numeric"
        obj = _np.asarray(values).view(cls)
                    
        if len(obj) == 0:
            _PhUI.i("Creating empty " + cls.__name__)
            
        obj._pyphysio = {
            cls._MT_SAMPLING_FREQ: sampling_freq,
            cls._MT_START_TIME: start_time if start_time is not None else 0,
            cls._MT_NATURE: signal_type
        }
        setattr(obj, "_mutated", False)
        return obj

    def __array_finalize__(self, obj):
        # __new__ called if obj is None
        if obj is not None and hasattr(obj, self._MT_INFO_ATTR):
            # The cache is not in MT_INFO_ATTR
            self._pyphysio = getattr(obj, self._MT_INFO_ATTR).copy()

    def __array_wrap__(self, out_arr, context=None):
        # Just call the parent's
        # noinspection PyArgumentList
        if isinstance(out_arr, Signal):
            return _np.ndarray.__array_wrap__(self, out_arr, context)
        else:
            return out_arr

    @property
    def ph(self):
        return self._pyphysio

    def clone(self):
        obj = self.copy()
        obj._pyphysio = copy.deepcopy(self.ph)
        return(obj)
    
    def is_multi(self):
        return(self.get_nchannels()>1)
    
    def get_values(self):
        return _np.asarray(self)
    
    def get_nchannels(self):
        if self.ndim>1:
            return(self.shape[1])
        else:
            return(1)
        
    def get_sampling_freq(self):
        return self.ph[self._MT_SAMPLING_FREQ]

    def set_sampling_freq(self, value):
        setattr(self, "_mutated", True)
        self.ph[self._MT_SAMPLING_FREQ] = value
    
    def get_start_time(self):
        return self.ph[self._MT_START_TIME]

    def set_start_time(self, value):
        setattr(self, "_mutated", True)
        self.ph[self._MT_START_TIME] = value    
    
    def get_signal_type(self):
        return self.ph[self._MT_NATURE]

    def set_signal_type(self, value):
        setattr(self, "_mutated", True)
        self.ph[self._MT_NATURE] = value
    
    def get_duration(self):
        return self.get_end_time() - self.get_start_time()
    
    def get_idx(self, time):
        idx = int((time - self.get_start_time()) * self.get_sampling_freq())
        if idx < 0:
            idx=0
        return(idx)
        
    @_abstract
    def clone_properties(self):
        pass
    
    @_abstract
    def get_times(self):
        pass
    
    @_abstract
    def get_end_time(self):
        pass

    @_abstract
    def get_iidx(self, time):
        pass

    @_abstract
    def get_time(self, idx):
        pass

    @_abstract
    def get_time_from_iidx(self, iidx):
        pass

    @_abstract
    def resample(self, fout, kind='linear'):
        pass

    @_abstract
    def segment_idx(self, t_start, t_stop=None):
        pass

    @_abstract
    def segment_iidx(self, t_start, t_stop=None):
        pass

    @_abstract
    def segment_time(self, t_start, t_stop=None):
        pass

    def plot(self, style="", vlines_height=1000):
        _xlabel("time")
        _ylabel(self.get_signal_type())
        _grid()
        if len(style) > 0 and style[0] == "|":
            return _vlines(self.get_times(), -vlines_height / 2, vlines_height / 2, style[1:])
        else:
            return _plot(self.get_times(), self.get_values(), style)

    @property
    def pickleable(self):
        """
        Returns a pickleable tuple of this Signal.
        :return: Tuple (Signal, ph dict).
        """
        return self, self.ph

    def to_pickle(self, path):
        """
        Saves this Signal into a pickle file.
        :param path: File system path to the file to write (create/overwrite).
        """
        from gzip import open
        from pickle import dump
        f = open(path, "wb")
        dump(self.pickleable, f, protocol=2)
        f.close()

#    def impute_nans(self):
#        self = ImputeNAN()(self)
        
    def __repr__(self):
        return "<signal: " + self.get_signal_type() + ", start_time: " + str(self.get_start_time()) + ">"

    def __getslice__(self, i, j):
        return self.segment_iidx(i, j)


class EvenlySignal(Signal):
    """
    Evenly spaced signal
    
    Attributes:
    -----------
    
    data : numpy.array
        Values of the signal
    sampling_freq : float, >0
        Sampling frequency
    start_time: float,
        Instant of signal start
    signal_type : str, default = ''
        Type of signal (e.g. 'ECG', 'EDA')
    """

    def clone_properties(self, new_values):
        x_new = EvenlySignal(new_values,
                             self.get_sampling_freq(),
                             self.get_start_time(),
                             self.get_signal_type())
        return(x_new)

    def get_times(self):
        return _np.arange(len(self)) / self.get_sampling_freq() + self.get_start_time()

    def get_end_time(self):
        return self.get_time(len(self) - 1) + 1. / self.get_sampling_freq()

    def get_iidx(self, time):
        return self.get_idx(time)

    def get_time(self, idx):
        return idx / self.get_sampling_freq() + self.get_start_time() if idx is not None else None

    def get_time_from_iidx(self, iidx):
        return self.get_time(iidx)

    def get_value_t(self, instant):
        values = self.get_values()
        nearest_idx = int(_np.round(self.get_sampling_freq() * (instant - self.get_start_time())))
        assert nearest_idx < len(self), "Required instant is after the end of the signal"  # return self[-1]
        assert nearest_idx >= 0, "Required instant is before the start of the signal"  # return self[0]
        
        return values[nearest_idx]
    
    def resample(self, fout, kind='linear'):
        """
        Resample a signal

        Parameters
        ----------
        fout : float
            The sampling frequency for resampling
        kind : str
            Method for interpolation: 'linear', 'nearest', 'zero', 'slinear', 'quadratic, 'cubic'

        Returns
        -------
        resampled_signal : EvenlySignal
            The resampled signal
        """

        ratio = self.get_sampling_freq() / fout

        if fout < self.get_sampling_freq() and ratio.is_integer():  # fast interpolation
            signal_out = self.get_values()[::int(ratio)]
        else:
            # The last sample is doubled to allow the new size to be correct
            indexes = _np.arange(len(self) + 1)
            indexes_out = _np.arange(len(self) * fout / self.get_sampling_freq()) * ratio
            self_l = _np.append(self, self[-1])

            if kind == 'cubic':
                tck = _interp.InterpolatedUnivariateSpline(indexes, self_l)
            else:
                tck = _interp.interp1d(indexes, self_l, kind=kind)
            signal_out = tck(indexes_out)

        return EvenlySignal(values=signal_out,
                            sampling_freq=fout,
                            signal_type=self.get_signal_type(),
                            start_time=self.get_start_time())

    # TRYME
    def segment_idx(self, idx_start, idx_stop=None):
        """
        Segment the signal given the indexes

        Parameters
        ----------
        idx_start : int or None
            The index of the start of the interval
        idx_stop : int or None
            The index of the end of the interval. By default is the length of the signal

        Returns
        -------
        portion : EvenlySignal
            The selected portion
        """
        return self.segment_iidx(idx_start, idx_stop)

    # TRYME
    def segment_iidx(self, iidx_start, iidx_stop=None):

        signal_values = self.get_values()

        if iidx_start is None:
            iidx_start = 0
        if iidx_stop is None:
            iidx_stop = len(self)

        values = signal_values[int(iidx_start):int(iidx_stop)]

        out_signal = self.clone_properties(values)
        out_signal.set_start_time(self.get_time(iidx_start))
        return out_signal

    # TRYME
    def segment_time(self, t_start, t_stop=None):
        """
        Segment the signal given a time interval

        Parameters
        ----------
        t_start : float
            The instant of the start of the interval
        t_stop : float 
            The instant of the end of the interval. By default is the end of the signal

        Returns
        -------
        portion : EvenlySignal
            The selected portion
        """

        return self.segment_idx(self.get_idx(t_start), self.get_idx(t_stop))

    def to_csv(self, filename, comment=''):
        values = self.get_values()
        times = self.get_times()
        header = self.get_signal_type() + ' \n' + 'Fsamp: ' + str(
            self.get_sampling_freq()) + '\n' + comment + '\nidx,time,value'

        _np.savetxt(filename, _np.c_[times, values], delimiter=',', header=header, comments='')

    def __repr__(self):
        return Signal.__repr__(self)[:-1] + " freq:" + str(self.get_sampling_freq()) + "Hz>\n" + self.view(
            _np.ndarray).__repr__()


class UnevenlySignal(Signal):
    """
    Unevenly spaced signal
    
    Attributes:
    -----------
    
    data : numpy.array
        Values of the signal
    sampling_freq : float, >0
        Sampling frequency
    start_time: float,
        Instant of signal start
    signal_type : str, default = ''
        Type of signal (e.g. 'ECG', 'EDA')
    
    
    x_values : numpy.array of int
        Instants, or indices when the values are measured.
    x_type : str
        Type of x values given.
        Can be 'indices' or 'instants'

    duration: float,
        Duration of the original EvenlySignal, if any. Duration is needed to have information about the duration of the
        last sample, if None the last sample will last 1. / fsamp.
    """

    _MT_X_INDICES = "x_values"
    _MT_DURATION = "duration"

    def __new__(cls, values, sampling_freq=1000, start_time=None, signal_type="", x_values=None, x_type='instants',
                duration=None):
        assert x_values is not None, "x_values are missing"
        assert x_type in ['indices', 'instants'], "x_type not in ['indices', 'instants']"
        x_values = _np.asarray(x_values)
        assert len(x_values) == len(values), "Length mismatch (y:%d vs. x:%d)" % (len(values), len(x_values))
        assert len(_np.where(_np.diff(x_values) <= 0)[0]) == 0, 'Given x_values are not strictly monotonic'

        
        if x_type == 'indices':
            # Keep indices, set start_time
            if start_time is None:
                start_time = 0
        else:
            
            # Get indices removing start_time
            if start_time is None:
                start_time = x_values[0]
            else:
                assert start_time <= x_values[0], "More than one sample at or before start_time"
            
            # WARN: limitation to 10 decimals due to workaround to prevent wrong cast flooring
            # (e.g. np.floor(0.29 * 100) == 28)
            x_values = _np.round((x_values - start_time) * sampling_freq, 10).astype(int)

        # adding 1/f cause end_time is exclusive 
        # Doesn't work when I put the last sample of a signal (we should add 1/fsamp to Signal.get_duration() too)
        min_duration = (x_values[-1] + 1.) / sampling_freq if len(x_values) > 0 else 0
        assert duration is None or duration >= min_duration, \
            "The specified duration is less than the one of the x_values"

        obj = Signal.__new__(cls, values=values,
                             sampling_freq=sampling_freq,
                             start_time=start_time,
                             signal_type=signal_type)

        if duration is None:
            duration = min_duration

        obj.ph[cls._MT_X_INDICES] = x_values
        obj.ph[cls._MT_DURATION] = duration
        return obj

    def clone_properties(self, new_values, new_x, new_x_type):
        x_new = UnevenlySignal(new_values,
                               self.get_sampling_freq(),
                               self.get_start_time(),
                               self.get_signal_type(),
                               new_x,
                               new_x_type)
        # TODO: test clone properties
        return(x_new)

    def get_duration(self):
        return self.ph[UnevenlySignal._MT_DURATION]

    def get_end_time(self):
        return self.get_start_time() + self.get_duration()

    def get_times(self):
        return self.ph[self._MT_X_INDICES] / self.get_sampling_freq() + self.get_start_time()

    def get_indices(self):
        return self.ph[self._MT_X_INDICES]

    def get_time(self, idx):
        return idx / self.get_sampling_freq() + self.get_start_time() if idx is not None else None

    def get_time_from_iidx(self, iidx):
        if len(self) == 0:
            return self.get_start_time()
        elif int(iidx) < len(self):
            return self.get_indices()[int(iidx)] / self.get_sampling_freq() + self.get_start_time()
        else:
            return self.get_time_from_iidx(-1)

    def get_iidx(self, time):
        return self.get_iidx_from_idx((time - self.get_start_time()) * self.get_sampling_freq())

    def get_iidx_from_idx(self, idx):
        if idx >= self.get_indices()[0]:
            return int(_np.searchsorted(self.get_indices(), idx))
        else:
            return None

    def to_csv(self, filename, comment=''):
        values = self.get_values()
        times = self.get_times()
        idxs = self.get_indices()
        header = self.get_signal_type() + ' \n' + 'Fsamp: ' + str(
            self.get_sampling_freq()) + '\n' + comment + '\nidx,time,value'

        _np.savetxt(filename, _np.c_[idxs, times, values], delimiter=',', header=header, comments='')

    def to_evenly(self, kind='cubic'):
        """
        Interpolate the UnevenlySignal to obtain an evenly spaced signal
        Parameters
        ----------
        kind : str
            Method for interpolation: 'linear', 'nearest', 'zero', 'slinear', 'quadratic, 'cubic'

        Returns
        -------
        interpolated_signal: ndarray
            The interpolated signal
        """

        assert kind != 'cubic' or len(self) > 3, "At least 4 samples needed for cubic interpolation"

        data_x = self.ph[self._MT_X_INDICES]  # From a constant freq range
        data_y = self.get_values()

        # Cubic if needed
        if kind == 'cubic':
            tck = _interp.InterpolatedUnivariateSpline(data_x, data_y)
        else:
            tck = _interp.interp1d(data_x, data_y, kind=kind)

        # Exclusive end, same x_value
        x_out = _np.arange(data_x[0], data_x[-1] + 1)
        sig_out = tck(x_out)

        # Init new signal
        sig_out = EvenlySignal(values=sig_out,
                               sampling_freq=self.get_sampling_freq(),
                               signal_type=self.get_signal_type(),
                               start_time=self.get_time_from_iidx(0))

        return sig_out

    def resample(self, fout, kind='linear'):
        return self.to_evenly(kind).resample(fout, kind)

    def segment_time(self, t_start, t_stop=None):
        """
        Segment the signal given a time interval

        Parameters
        ----------
        t_start : float
            The instant of the start of the interval
        t_stop : float 
            The instant of the end of the interval. By default is the end of the signal

        Returns
        -------
        portion : UnvenlySignal
            The selected portion
        """

        return self.segment_idx(self.get_idx(t_start) if t_start is not None else None,
                                self.get_idx(t_stop) if t_stop is not None else None)

    def segment_idx(self, idx_start, idx_stop=None):
        """
        Segment the signal given the indexes

        Parameters
        ----------
        idx_start : int
            The index of the start of the interval
        idx_stop : float
            The index of the end of the interval. By default is the end of the signal

        Returns
        -------
        portion : UnvenlySignal
            The selected portion
        """
        if idx_start is None:
            idx_start = 0
        if idx_stop is None:
            idx_stop = self.get_indices()[-1]

        iib = self.get_iidx_from_idx(idx_start)
        iie = self.get_iidx_from_idx(idx_stop)

        if iib is None and iie is None:
            iidx_start = iidx_stop = idx_start = idx_stop = 0
        else:
            iidx_start = int(iib) if iib is not None else 0
            iidx_stop = int(iie) if iie is not None else -1

        return UnevenlySignal(values=self.get_values()[iidx_start:iidx_stop],
                              x_values=self.get_indices()[iidx_start:iidx_stop] - idx_start,
                              sampling_freq=self.get_sampling_freq(),
                              signal_type=self.get_signal_type(),
                              start_time=self.get_time(idx_start),
                              x_type='indices',
                              duration=(idx_stop - idx_start) / self.get_sampling_freq())

    def segment_iidx(self, iidx_start, iidx_stop=None):
        """
        Segment the signal given the inner indexes

        Parameters
        ----------
        iidx_start : int
            The index of the start of the interval
        iidx_stop : float
            The index of the end of the interval. By default is the end of the signal

        Returns
        -------
        portion : UnvenlySignal
            The selected portion
        """
        if iidx_stop is None:
            iidx_stop = len(self)
        if iidx_start is None:
            iidx_start = 0
        if iidx_stop < len(self):
            idx_stop = self.get_indices()[int(iidx_stop)]
        else:
            idx_stop = self.get_indices()[-1] + 1
        idx_start = self.get_indices()[int(iidx_start)]

        return UnevenlySignal(values=self.get_values()[int(iidx_start):int(iidx_stop)],
                              x_values=self.get_indices()[int(iidx_start):int(iidx_stop)]
                              - self.get_indices()[int(iidx_start)],
                              sampling_freq=self.get_sampling_freq(),
                              signal_type=self.get_signal_type(),
                              start_time=self.get_time_from_iidx(iidx_start),
                              x_type='indices',
                              duration=(idx_stop - idx_start) / self.get_sampling_freq())

    def __repr__(self):
        return Signal.__repr__(self)[:-1] + " time resolution:" + str(1 / self.get_sampling_freq()) + "s>\n" + \
               self.get_values().__repr__() + " Times\n:" + self.get_times().__repr__()


class MultiEvenly(EvenlySignal):
    
    def __new__(cls, values, sampling_freq, start_time=None, signal_type='raw', info = {}):
        assert sampling_freq > 0, "The sampling frequency cannot be zero or negative"
        assert start_time is None or isinstance(start_time, _Number), "Start time is not numeric"
        obj = Signal.__new__(cls, values=values, sampling_freq=sampling_freq, start_time=start_time, signal_type=signal_type)
        
        return obj

    def clone_properties(self, new_values):
        x_new = MultiEvenly(new_values,
                            self.get_sampling_freq(),
                            self.get_start_time(),
                            self.get_signal_type())
        return(x_new)
    
    
    def set_start_time(self, value):
        setattr(self, "_mutated", True)
        self.ph[self._MT_START_TIME] = value  
        stim = self.get_stim()
        stim.set_start_time(value)
        
    def get_channel(self, i_ch):
        ch_values = self.get_values()[:,i_ch]
        return(EvenlySignal(ch_values, self.get_sampling_freq(), self.get_start_time(), self.get_signal_type()))
        
    def resample(self, fout, kind='linear'):
        """
        Resample a signal

        Parameters
        ----------
        fout : float
            The sampling frequency for resampling
        kind : str
            Method for interpolation: 'linear', 'nearest', 'zero', 'slinear', 'quadratic, 'cubic'

        Returns
        -------
        resampled_signal : EvenlySignal
            The resampled signal
        """

        ratio = self.get_sampling_freq() / fout

        if fout < self.get_sampling_freq() and ratio.is_integer():  # fast interpolation
            values_out = self.get_values()[::int(ratio),:]
        else:
            indexes = _np.arange(len(self) + 1)
            indexes_out = _np.arange(len(self) * fout / self.get_sampling_freq()) * ratio
            
            values = self.get_values()
            values = _np.vstack([values, values[-1,:]])
            tck = _interp.interp1d(indexes, values, kind=kind, axis=0)
            
            values_out = tck(indexes_out)
    
        signal_out = self.clone_properties(values_out)
        signal_out.set_sampling_freq(fout)
        return(signal_out)
        

    # TRYME
    def segment_iidx(self, iidx_start, iidx_stop=None):

        signal_values = self.get_values()

        if iidx_start is None:
            iidx_start = 0
        if iidx_stop is None:
            iidx_stop = len(self)

        values = signal_values[int(iidx_start):int(iidx_stop),:]

        out_signal = self.clone_properties(values)
        out_stim = self.get_stim().clone()
        out_stim = out_stim.segment_iidx(iidx_start, iidx_stop)
        out_signal.set_stim(out_stim)
        
        out_signal.set_start_time(self.get_time(iidx_start))
        return out_signal
    
    
    def plot(self, style=""):
        _grid()
        n_ch = self.get_nchannels()
    
        t_start = self.get_start_time()
        
        n_rows = int(_np.ceil(n_ch/4))
        ax1 = _subplot(n_rows, 4, 1)
        for i_ch in range(n_ch):
            _subplot(n_rows, 4, i_ch+1, sharex=ax1)
            self.get_channel(i_ch).plot(style)
            _ylabel(i_ch)
            _vlines(t_start, _np.nanmin(self.get_channel(i_ch)), _np.nanmax(self.get_channel(i_ch)), 'k')
            _grid()
        _xlim(self.get_start_time(), self.get_end_time())
        _tight_layout()
        _subplots_adjust(top=0.9, bottom=0.01, left=0.05, right=0.95, hspace=0.3, wspace=0.25)

    def to_csv(self, filename, comment='', fmt = '%.6f'): 
        values = self.get_values()
        times = self.get_times()
        header = self.get_signal_type() + ' \n' + 'Fsamp: ' + str(self.get_sampling_freq()) + '\n' + comment + '\nidx,time'+''.join([f',ch{x}' for x in range(self.get_nchannels())])
        _np.savetxt(filename, _np.c_[_np.arange(len(times)), times, values], delimiter=',', header=header, comments='', fmt = fmt)

    def __repr__(self):
        return f"<start_time: {self.get_start_time()}> freq:  {self.get_sampling_freq()} Hz> {self.view(_np.ndarray).__repr__()}"

    def __getslice__(self, i, j):
        return self.segment_iidx(i, j)


# coding=utf-8
import numpy as _np
from ..Utility import PhUI as _PhUI, abstractmethod as _abstract
from ..BaseSegmentation import SegmentsGenerator, Segment
from ..Signal import Signal as _Signal

__author__ = 'AleB'


class _SegmentsWithLabelSignal(SegmentsGenerator):
    # Assumed: label signal extended over the end by holding the value

    def __init__(self, drop_cut=True, drop_mixed=True, **kwargs):
        super(_SegmentsWithLabelSignal, self).__init__(drop_cut=drop_cut, drop_mixed=drop_mixed, **kwargs)
        self._labsig = None

    @_abstract
    def init_segmentation(self):
        pass

    def next_segment(self):
        if self._signal is None:
            _PhUI.w("Can't preview the segments without a signal here. Use the syntax " +
                    FixedSegments.__name__ + "(**params)(signal)")
            raise StopIteration()

        b, e, label = self.next_segment_mix_labels()
        s = Segment(b, e, label, self._signal)
        return s

    @_abstract
    def next_times(self):
        pass

    def check_drop_and_range(self, s, b, e):
        drop = True, None, None

        # signal segment bounds

        # full under-range (empty) or full over-range (empty)
        if e < s.get_start_time() or b >= s.get_end_time():
            return drop

        # part before start: mixed and shorter (as partially before the first label's begin)
        if b < s.get_start_time():
            if self._params['drop_mixed'] or self._params['drop_cut']:
                # goto next segment (drop)
                return drop
            else:
                # cut to start
                b = s.get_start_time()

        # part after end: shorter but not mixed (as half after the last label's end)
        if e > s.get_end_time():
            if self._params['drop_cut']:
                # goto next segment (drop)
                return drop
            else:
                # cut to start
                e = s.get_end_time()

        # Don't drop, inclusive begin time, exclusive end time
        return False, b, e

    def next_segment_mix_labels(self):
        label = b = e = None
        while True:
            # break    ==> keep
            # continue ==> drop

            b, e = self.next_times()


            drop, b, e = self.check_drop_and_range(self._signal, b, e)

            if drop:
                continue

            if not isinstance(self._labsig, _Signal):
                label = None
            else:

                drop, _, _ = self.check_drop_and_range(self._labsig, b, e)

                if drop:
                    # partially or completely out of labsig range and have to drop it
                    label = None
                    continue

                # labels segment bounds, may be < 0 (None < 0)
                first = self._labsig.get_iidx(b)
                last = self._labsig.get_iidx(e)
                
                if first == last:
                    last += 1

                lab_seg = self._labsig.segment_iidx(first, last)
                lab_first = lab_seg[0]

                if len(lab_seg) == 1 or (lab_seg[:1:-1] == lab_first).all(): ###[AB]
                    label = lab_first
                else:
                    if self._params['drop_mixed']:
                        continue
                    else:
                        label = None
            break

        return b, e, label


class RandomFixedSegments(_SegmentsWithLabelSignal):
    """
    Fixed length segments iterator, at random start timestamps, specifying step and width in seconds.

    A label signal from which to
    take labels can be specified.

    Parameters
    ----------
    width : float, >0
        time distance between subsequent segments.
    N : int, >0
        number of segments to be extracted.

    Optional parameters
    -------------------
    labels : array
        Signal of the labels
    drop_mixed : bool, default=True
        In case labels is specified, whether to drop segments with more than one label, if False the label of such
         segments is set to None.
    drop_cut : bool, default=True
        Whether to drop segments that are shorter due to the crossing of the signal end.
    """

    def __init__(self, N, width, start=None, stop = None, labels=None, drop_mixed=True, drop_cut=True, **kwargs):
        super(RandomFixedSegments, self).__init__(N=N, width=width, start=start, stop=stop, labels=labels, drop_cut=drop_cut,
                                            drop_mixed=drop_mixed, **kwargs)
        assert N > 0
        assert width > 0
        assert start is None or isinstance(start, float)
        assert stop is None or isinstance(stop, float)
        assert labels is None or isinstance(labels, _Signal),\
            "The parameter 'labels' should be a Signal."
        self._width = None
        self._tst = None
        self._tsp = None
        self._i = None

    def init_segmentation(self):
        self._N = self._params["N"]
        w = self._params["width"]
        self._width = w 
        self._labsig = self._params["labels"]
        self._tst = self._params["start"]
        self._tsp = self._params["stop"]
        
        #generate random start instants
        tst = self._signal.get_start_time() if self._tst is None else self._tst
        tsp = self._signal.get_end_time() if self._tsp is None else self._tsp
        tsp = tsp - w
        self._tst_randoms = _np.random.uniform(tst, tsp, self._N)
        
    def next_times(self):
        if self._i is None:
            self._i = 0
        
        if self._i < len(self._tst_randoms):
            b = self._tst_randoms[self._i]
            e = b + self._width
            self._i += 1
            return b, e
        else:
            raise StopIteration()
            

class FixedSegments(_SegmentsWithLabelSignal):
    """
    Fixed length segments iterator, specifying step and width in seconds.

    A label signal from which to
    take labels can be specified.

    Parameters
    ----------
    step : float, >0
        time distance between subsequent segments.

    Optional parameters
    -------------------
    width : float, >0, default=step
        time distance between subsequent segments.
    start : float
        start time of the first segment
    labels : array
        Signal of the labels
    drop_mixed : bool, default=True
        In case labels is specified, whether to drop segments with more than one label, if False the label of such
         segments is set to None.
    drop_cut : bool, default=True
        Whether to drop segments that are shorter due to the crossing of the signal end.
    """

    def __init__(self, step, width=None, start=None, labels=None, drop_mixed=True, drop_cut=True, **kwargs):
        super(FixedSegments, self).__init__(step=step, width=width, start=start, labels=labels, drop_cut=drop_cut,
                                            drop_mixed=drop_mixed, **kwargs)
        assert step > 0
        assert width is None or width > 0
        assert start is None or start > 0
        assert labels is None or isinstance(labels, _Signal),\
            "The parameter 'labels' should be a Signal."
        self._step = None
        self._width = None
        self._t = None

    def init_segmentation(self):
        self._step = self._params["step"]
        w = self._params["width"]
        self._width = w if w is not None else self._step
        self._labsig = self._params["labels"]
        self._t = self._params["start"]

    def next_times(self):
        if self._t is None:
            self._t = self._signal.get_start_time()
        b = self._t
        self._t += self._step
        e = b + self._width
        if b >= self._signal.get_end_time():
            raise StopIteration()
        return b, e

class CustomSegments(_SegmentsWithLabelSignal):
    """
    Custom segments iterator, specifying an array of begin times and an array of end times.

    Parameters
    ----------
    begins : array or list
        Array of the begin times of the segments to return.
    ends : array or list
        Array of the end times of the segments to return, of the same length of 'begins'.

    Optional parameters
    -------------------
    labels : array or list
        Signal of the labels
    drop_mixed : bool, default=True
        In case labels is specified, weather to drop segments with more than one label, if False the label of such
         segments is set to None.
    drop_cut : bool, default=True
        Weather to drop segments that are shorter due to the crossing of the signal end.
    """

    def __init__(self, begins, ends, labels=None, drop_mixed=True, drop_cut=True, **kwargs):
        super(CustomSegments, self).__init__(begins=begins, ends=ends, labels=labels, drop_cut=drop_cut,
                                             drop_mixed=drop_mixed, **kwargs)
        assert len(begins) == len(ends), "The number of begins has to be equal to the number of ends :)"
        assert labels is None or isinstance(labels, _Signal),\
            "The parameter 'labels' should be an Signal."
        self._i = None
        self._b = None
        self._e = None

    def init_segmentation(self):
        self._i = -1
        self._b = self._params['begins']
        self._e = self._params['ends']
        self._labsig = self._params["labels"]

    def next_times(self):
        self._i += 1
        if self._i < len(self._b):
            return self._b[self._i], self._e[self._i]
        else:
            raise StopIteration()


class LabelSegments(_SegmentsWithLabelSignal):
    """
    Generates a list of segments from a label signal, allowing to collapse subsequent equal samples.

    Parameters
    ----------
    labels : array or list
        Signal of the labels

    Optional parameters
    -------------------
    drop_mixed : bool, default=True
        In case labels is specified, weather to drop segments with more than one label, if False the label of such
         segments is set to None.
    drop_cut : bool, default=True
        Weather to drop segments that are shorter due to the crossing of the signal end.
    """

    def __init__(self, labels, drop_mixed=True, drop_cut=True, **kwargs):
        super(LabelSegments, self).__init__(labels=labels, drop_mixed=drop_mixed, drop_cut=drop_cut, **kwargs)
        assert labels is None or isinstance(labels, _Signal),\
            "The parameter 'labels' should be an Signal."
        self._i = None
        self._labsig = None

    def init_segmentation(self):
        self._i = 0
        self._labsig = self._params['labels']

    def next_times(self):
        if self._i >= len(self._labsig):
            raise StopIteration()
        end = self._i
        while end < len(self._labsig) and self._labsig[self._i] == self._labsig[end]:
            end += 1
        
        b = self._labsig.get_time_from_iidx(self._i)
        e = self._labsig.get_time_from_iidx(end)
        self._i = end
        return b, e



# coding=utf-8
from pyphysio.BaseAlgorithm import Algorithm
from abc import ABCMeta as _ABCMeta
__author__ = 'AleB'


class Indicator(Algorithm):
    """
    Algorithms that take as input a signal and return a scalar value.
    """
    __metaclass__ = _ABCMeta
    pass



# coding=utf-8
from __future__ import division

import numpy as _np

from ..BaseIndicator import Indicator as _Indicator
from ..tools.Tools import Diff as _Diff
from ..Signal import EvenlySignal as _EvenlySignal, Signal as _Signal


__author__ = 'AleB'


class Mean(_Indicator):
    """
    Compute the arithmetic mean of the signal, ignoring any NaNs.
    """
    def __init__(self, **kwargs):
        _Indicator.__init__(self, **kwargs)

    @classmethod
    def algorithm(cls, data, params):
        return _np.nanmean(data.get_values())


class Min(_Indicator):
    """
    Return minimum of the signal, ignoring any NaNs.
    """
    def __init__(self, **kwargs):
        _Indicator.__init__(self, **kwargs)

    @classmethod
    def algorithm(cls, data, params):
        return _np.nanmin(data.get_values())


class Max(_Indicator):
    """
    Return maximum of the signal, ignoring any NaNs.
    """
    def __init__(self, **kwargs):
        _Indicator.__init__(self, **kwargs)

    @classmethod
    def algorithm(cls, data, params):
        return _np.nanmax(data.get_values())


class Range(_Indicator):
    """
    Compute the range of the signal, ignoring any NaNs.
    """
    def __init__(self, **kwargs):
        _Indicator.__init__(self, **kwargs)

    @classmethod
    def algorithm(cls, data, params):
        return Max()(data) - Min()(data)


class Median(_Indicator):
    """
    Compute the median of the signal, ignoring any NaNs.
    """
    def __init__(self, **kwargs):
        _Indicator.__init__(self, **kwargs)

    @classmethod
    def algorithm(cls, data, params):
        return _np.median(data.get_values())


class StDev(_Indicator):
    """
    Computes the standard deviation of the signal, ignoring any NaNs.
    """
    def __init__(self, **kwargs):
        _Indicator.__init__(self, **kwargs)

    @classmethod
    def algorithm(cls, data, params):
        return _np.nanstd(data.get_values())


class Sum(_Indicator):
    """
    Computes the sum of the values in the signal, treating Not a Numbers (NaNs) as zero.
    """
    def __init__(self, **kwargs):
        _Indicator.__init__(self, **kwargs)

    @classmethod
    def algorithm(cls, data, params):
        return _np.nansum(data.get_values())


class AUC(_Indicator):
    """
    Computes the Area Under the Curve of the signal, treating Not a Numbers (NaNs) as zero.
    """
    def __init__(self, **kwargs):
        _Indicator.__init__(self, **kwargs)

    @classmethod
    def algorithm(cls, signal, params):
        if isinstance(signal, _Signal) and not isinstance(signal, _EvenlySignal):
            cls.warn('Calculating Area Under the Curve of an Unevenly signal!')
        fsamp = signal.get_sampling_freq()
        return (1. / fsamp) * Sum()(signal)
    
class DetrendedAUC(_Indicator):
    """
    Computes the Area Under the Curve of the signal, treating Not a Numbers (NaNs) as zero.
    """
    def __init__(self, **kwargs):
        _Indicator.__init__(self, **kwargs)

    @classmethod
    def algorithm(cls, signal, params):
        if isinstance(signal, _Signal) and not isinstance(signal, _EvenlySignal):
            cls.warn('Calculating Area Under the Curve of an Unevenly signal!')
        fsamp = signal.get_sampling_freq()
        
        #detrend
        t_signal = signal.get_times()
        intercept = signal[0]
        coeff = (signal[-1] - signal[0]) / signal.get_duration()
        baseline = coeff*(t_signal - t_signal[0]) + intercept
        
        signal_ = signal - baseline
        return (1. / fsamp) * Sum()(signal_)


class RMSSD(_Indicator):
    """
    Compute the square root of the mean of the squared 1st order discrete differences.
    """
    def __init__(self, **kwargs):
        _Indicator.__init__(self, **kwargs)

    @classmethod
    def algorithm(cls, signal, params):
        diff = _Diff()(signal)
        return _np.sqrt(_np.mean(_np.power(diff.get_values(), 2)))


class SDSD(_Indicator):
    """
    Calculate the standard deviation of the 1st order discrete differences.
    """
    def __init__(self, **kwargs):
        _Indicator.__init__(self, **kwargs)

    @classmethod
    def algorithm(cls, signal, params):
        diff = _Diff()(signal)
        return StDev()(diff)

# TODO: FIX Histogram missing
class Triang(_Indicator):
    """
    Computes the HRV triangular index.
    """
    def __init__(self, **kwargs):
        _Indicator.__init__(self, **kwargs)

    @classmethod
    def algorithm(cls, data, params):
        step = 1000. / 128
        min_ibi = _np.min(data)
        max_ibi = _np.max(data)
        if (max_ibi - min_ibi) / step + 1 < 10:
            cls.warn("len(bins) < 10")
            return _np.nan
        else:
            bins = _np.arange(min_ibi, max_ibi, step)
            h, b = Histogram(histogram_bins=bins)(data)
            return len(data) / _np.max(h)


class TINN(_Indicator):
    """
    Computes the triangular interpolation of NN interval histogram.
    """
    def __init__(self, **kwargs):
        _Indicator.__init__(self, **kwargs)

    @classmethod
    def algorithm(cls, data, params):
        step = 1000. / 128
        min_ibi = _np.min(data)
        max_ibi = _np.max(data)
        if (max_ibi - min_ibi) / step + 1 < 10:
            cls.warn("len(bins) < 10")
            return _np.nan
        else:
            bins = _np.arange(min_ibi, max_ibi, step)
            h, b = Histogram(histogram_bins=bins)(data)
            max_h = _np.max(h)
            hist_left = _np.array(h[0:_np.argmax(h)])
            ll = len(hist_left)
            hist_right = _np.array(h[_np.argmax(h):])
            rl = len(hist_right)
            y_left = _np.array(_np.linspace(0, max_h, ll))

            minx = _np.Inf
            pos = 0
            for i in range(1, len(hist_left) - 1):
                curr_min = _np.sum((hist_left - y_left) ** 2)
                if curr_min < minx:
                    minx = curr_min
                    pos = i
                y_left[i] = 0
                y_left[i + 1:] = _np.linspace(0, max_h, ll - i - 1)

            n = b[pos - 1]

            y_right = _np.array(_np.linspace(max_h, 0, rl))
            minx = _np.Inf
            pos = 0
            for i in range(rl - 1, 1, -1):
                curr_min = _np.sum((hist_right - y_right) ** 2)
                if curr_min < minx:
                    minx = curr_min
                    pos = i
                y_right[i - 1] = 0
                y_right[0:i - 2] = _np.linspace(max_h, 0, i - 2)

            m = b[_np.argmax(h) + pos + 1]
            return m - n



# coding=utf-8
from __future__ import division
import numpy as _np
import scipy.stats as _stats
from scipy.signal import gaussian as _gaussian, filtfilt as _filtfilt, filter_design as _filter_design, \
    deconvolve as _deconvolve, firwin as _firwin, convolve as _convolve
from matplotlib.pyplot import plot as _plot
from ..BaseFilter import Filter as _Filter
from ..Signal import EvenlySignal as _EvenlySignal, UnevenlySignal as _UnevenlySignal
from ..Utility import abstractmethod as _abstract
from ..tools.Tools import SignalRange
from collections import Sequence
__author__ = 'AleB'


class Normalize(_Filter):
    """
    Normalized the input signal using the general formula: ( signal - BIAS ) / RANGE

    Parameters
    -------------------
    norm_method : 
        Method for the normalization. Available methods are:
    * 'mean' - remove the mean [ BIAS = mean(signal); RANGE = 1 ]
    * 'standard' - standardization [ BIAS = mean(signal); RANGE = std(signal) ]
    * 'min' - remove the minimum [ BIAS = min(signal); RANGE = 1 ]
    * 'maxmin' - maxmin normalization [ BIAS = min(signal); RANGE = ( max(signal) - min(signal ) ]
    * 'custom' - custom, bias and range are manually defined [ BIAS = bias, RANGE = range ]
    
    norm_bias : float, default = 0
        Bias for custom normalization
    norm_range : float, !=0, default = 1
        Range for custom normalization

    Returns
    -------
    signal: 
        The normalized signal. 

    """

    def __init__(self, norm_method='standard', norm_bias=0, norm_range=1):
        assert norm_method in ['mean', 'standard', 'min', 'maxmin', 'custom'],\
            "norm_method must be one of 'mean', 'standard', 'min', 'maxmin', 'custom'"
        if norm_method == "custom":
            assert norm_range != 0, "norm_range must not be zero"
        _Filter.__init__(self, norm_method=norm_method, norm_bias=norm_bias, norm_range=norm_range)

    @classmethod
    def algorithm(cls, signal, params):
        from ..indicators.TimeDomain import Mean as _Mean, StDev as _StDev

        method = params['norm_method']
        if method == "mean":
            return signal - _Mean()(signal)
        elif method == "standard":
            return (signal - _Mean()(signal)) / _StDev()(signal)
        elif method == "min":
            return signal - _np.min(signal)
        elif method == "maxmin":
            return (signal - _np.min(signal)) / (_np.max(signal) - _np.min(signal))
        elif method == "custom":
            return (signal - params['norm_bias']) / params['norm_range']



class IIRFilter(_Filter):
    """
    Filter the input signal using an Infinite Impulse Response filter.

    Parameters
    ----------
    fp : list or float
        The pass frequencies
    fs : list or float
        The stop frequencies
    
    Optional parameters
    -------------------
    loss : float, >0, default = 0.1
        Loss tolerance in the pass band
    att : float, >0, default = 40
        Minimum attenuation required in the stop band.
    ftype : str, default = 'butter'
        Type of filter. Available types: 'butter', 'cheby1', 'cheby2', 'ellip', 'bessel'

    Returns
    -------
    signal : EvenlySignal
        Filtered signal

    Notes
    -----
    This is a wrapper of *scipy.signal.filter_design.iirdesign*. Refer to `scipy.signal.filter_design.iirdesign`
    for additional information
    """

    def __init__(self, fp, fs, loss=.1, att=40, ftype='butter'):
        assert loss > 0, "Loss value should be positive"
        assert att > 0, "Attenuation value should be positive"
        assert att > loss, "Attenuation value should be greater than loss value"
        assert ftype in ['butter', 'cheby1', 'cheby2', 'ellip', 'bessel'],\
            "Filter type must be in ['butter', 'cheby1', 'cheby2', 'ellip', 'bessel']"
        _Filter.__init__(self, fp=fp, fs=fs, loss=loss, att=att, ftype=ftype)

    @classmethod
    def algorithm(cls, signal, params):
        fsamp = signal.get_sampling_freq()
        fp, fs, loss, att, ftype = params["fp"], params["fs"], params["loss"], params["att"], params["ftype"]

        if isinstance(signal, _UnevenlySignal):
            cls.warn('Filtering Unevenly signal is undefined. Returning original signal.')
            return signal

        nyq = 0.5 * fsamp
        fp = _np.array(fp)
        fs = _np.array(fs)

        wp = fp / nyq
        ws = fs / nyq
        # noinspection PyTupleAssignmentBalance
        b, a = _filter_design.iirdesign(wp, ws, loss, att, ftype=ftype, output="ba")

        sig_filtered = signal.clone_properties(_filtfilt(b, a, signal.get_values()))

        if _np.isnan(sig_filtered[0]):
            cls.warn('Filter parameters allow no solution. Returning original signal.')
            return signal
        else:
            return sig_filtered

    @_abstract
    def plot(self):
        pass

class FIRFilter(_Filter):
    """
    Filter the input signal using a Finite Impulse Response filter.

    Parameters
    ----------
    fp : list or float
        The pass frequencies
    fs : list or float
        The stop frequencies
    
    Optional parameters
    -------------------
    loss : float, >0, default = 0.1
        Loss tolerance in the pass band
    att : float, >0, default = 40
        Minimum attenuation required in the stop band.
    wtype : str, default = 'hamming'
        Type of filter. Available types: 'hamming'

    Returns
    -------
    signal : EvenlySignal
        Filtered signal

    Notes
    -----
    This is a wrapper of *scipy.signal.firwin*. Refer to `scipy.signal.firwin`
    for additional information
    """

    def __init__(self, fp, fs, loss=0.1, att=40, wtype='hamming'):
        assert loss > 0, "Loss value should be positive"
        assert att > 0, "Attenuation value should be positive"
        assert att > loss, "Attenuation value should be greater than loss value"
        assert wtype in ['hamming'],\
            "Window type must be in ['hamming']"
        _Filter.__init__(self, fp=fp, fs=fs, loss=loss, att=att, wtype=wtype)

    @classmethod
    def algorithm(cls, signal, params):
        fsamp = signal.get_sampling_freq()
        fp, fs, loss, att, wtype = params["fp"], params["fs"], params["loss"], params["att"], params["wtype"]

        if isinstance(signal, _UnevenlySignal):
            cls.warn('Filtering Unevenly signal is undefined. Returning original signal.')
            return signal

        fp = _np.array(fp)
        fs = _np.array(fs)
        
        if att>0:
            att = -att
        d1 = 10**(loss/10)
        d2 = 10**(att/10)
        Dsamp = _np.min(abs(fs-fp))/fsamp
        

        # from https://dsp.stackexchange.com/questions/31066/how-many-taps-does-an-fir-filter-need
        N = int(2/3*_np.log10(1/(10*d1*d2))*fsamp/Dsamp)
                
        pass_zero=True
                  
        if isinstance(fp, Sequence):
            if fp[0]>fs[0]:
                pass_zero=False
        else:    
            if fp[0]>fs[0]:
                pass_zero=False
        
            
        
        nyq = 0.5 * fsamp
        fp = _np.array(fp)
        wp = fp / nyq
        
        if N%2 ==0:
            N+=1
        b = _firwin(N, wp, width=Dsamp, window=wtype, pass_zero=pass_zero)
        sig_filtered = signal.clone_properties(_convolve(signal.get_values(), b, mode='same'))

        if _np.isnan(sig_filtered[0]):
            cls.warn('Filter parameters allow no solution. Returning original signal.')
            return signal
        else:
            return sig_filtered

    @_abstract
    def plot(self):
        pass

class KalmanFilter(_Filter):
    def __init__(self, R, ratio=1, win_len=1, win_step=0.5):
        assert R > 0, "R should be positive"
        if ratio is not None:
            assert ratio > 1, "ratio should be >1"
        assert win_len > 0, "Window length value should be positive"
        assert win_step > 0, "Window step value should be positive"
        
        _Filter.__init__(self, R=R, ratio=ratio, win_len=win_len, win_step=win_step)
        
    @classmethod
    def algorithm(cls, signal, params):
        R = params['R']
        ratio = params['ratio']
        win_len = params['win_len']
        win_step = params['win_step']
        
        sz = len(signal)
        
        rr = SignalRange(win_len, win_step)(signal)
        Q = _np.nanmedian(rr)/ratio
            
        P = 1
        
        x_out = signal.get_values().copy()
        for k in range(1,sz):
                x_ = x_out[k-1]
                P_ = P + Q
            
                # measurement update
                K = P_ / (P_ + R)
                x_out[k] = x_ + K * (x_out[k] - x_)
                P = (1 - K ) * P_

        x_out = signal.clone_properties(x_out)
        return(x_out)

############
class ImputeNAN(_Filter):
    def __init__(self, win_len=5, allnan='nan'):
        assert win_len>0, "win_len should be >0"
        assert allnan in ['zeros', 'nan']
        _Filter.__init__(self, win_len = win_len, allnan=allnan)
        
    @classmethod
    def algorithm(cls, signal, params):
        def group_consecutives(vals, step=1):
            """Return list of consecutive lists of numbers from vals (number list)."""
            run = []
            result = [run]
            expect = None
            for v in vals:
                if (v == expect) or (expect is None):
                    run.append(v)
                else:
                    run = [v]
                    result.append(run)
                expect = v + step
            return result

        #%
        win_len = params['win_len']*signal.get_sampling_freq()
        allnan = params['allnan']
        
        s = signal.get_values().copy()
        if _np.isnan(s).all():
            if allnan == 'nan':
                return(signal)
            else:
                s = _np.zeros_like(s)
                s_out = signal.clone_properties(s)
                return(s_out)
        
        idx_nan = _np.where(_np.isnan(s))[0]
        segments = group_consecutives(idx_nan)

        #%
        if len(segments[0])>=1:
            for i_seg, SEG in enumerate(segments):
                idx_st = SEG[0]
                idx_sp = SEG[-1]
                idx_win_pre = _np.arange(-int(win_len/2), 0, 1)+idx_st
                idx_win_pre = idx_win_pre[_np.where(idx_win_pre>0)[0]] #not before signal start

                STD = []
                if len(idx_win_pre)>=3:
                    STD.append(_np.nanstd(s[idx_win_pre]))
                
                idx_win_post = _np.arange(0, int(win_len/2))+idx_sp+1
                idx_win_post = idx_win_post[_np.where(idx_win_post<len(s))[0]]
                
                if len(idx_win_post)>=3:
                    STD.append(_np.nanstd(s[idx_win_post]))
                
                if len(STD)>0 and not (_np.isnan(STD).all()):
                    STD = _np.nanmin(STD)
                else:
                    STD = 0
                    
                idx_win = _np.hstack([idx_win_pre, idx_win_post]).astype(int)
                idx_win = idx_win[_np.where(~_np.isnan(s[idx_win]))[0]] # remove nans
                
                if len(idx_win)>3:
                    R = _stats.linregress(idx_win, s[idx_win])
                    s_nan = _np.array(SEG)*R[0]+R[1] + _np.random.normal(scale=STD, size = len(SEG))
                else:
                    s_nan = _np.nanmean(s)*_np.ones(len(SEG))
                s[SEG] = s_nan
        
        signal_out = signal.clone_properties(s)
        return(signal_out)


class RemoveSpikes(_Filter):
    def __init__(self, K=2, N=1, dilate=0, D=0.95, method='step'):
        assert K > 0, "K should be positive"
        assert isinstance(N, int) and N>0, "N value not valid"
        assert dilate>=0, "dilate should be >= 0.0"
        assert D>=0, "D should be >= 0.0"
        assert method in ['linear', 'step']
        _Filter.__init__(self, K=K, N=N, dilate=dilate, D=D, method=method)
    
    @classmethod
    def algorithm(cls, signal, params):
        K = params['K']
        N = params['N']
        dilate = params['dilate']
        D = params['D']
        method = params['method']
        fs = signal.get_sampling_freq()
        
        sig_diff = abs(signal[N:] - signal[:-N])
        ds_mean = _np.nanmean(sig_diff)
        
        idx_spikes = _np.where(sig_diff>K*ds_mean)[0]+N//2
        spikes = _np.zeros(len(signal))
        spikes[idx_spikes] = 1
        win = _np.ones(1+int(2*dilate*fs))
        spikes = _np.convolve(spikes, win, 'same')
        idx_spikes = _np.where(spikes>0)[0]
        
        x_out = signal.get_values().copy()
        
        #TODO add linear connector method
        if method == 'linear':
            diff_idx_spikes = _np.diff(idx_spikes)
            new_spike = _np.where(diff_idx_spikes > 1)[0] + 1
            new_spike = _np.r_[0, new_spike, -1]
            for I in range(len(new_spike)-1):
                IDX_START = idx_spikes[new_spike[I]] -1
                IDX_STOP = idx_spikes[new_spike[I+1]-1] +1
                
                L = IDX_STOP - IDX_START + 1
                x_start = x_out[IDX_START]
                x_stop = x_out[IDX_STOP]
                coefficient = (x_stop - x_start)/ L
                
                x_out[IDX_START:IDX_STOP+1] = coefficient*_np.arange(L) + x_start
        else:
                
            for IDX in idx_spikes:
                delta = x_out[IDX] - x_out[IDX-1]
                x_out[IDX:] = x_out[IDX:] - D*delta
        x_out = signal.clone_properties(x_out)
        return(x_out)

class DenoiseEDA(_Filter):
    """
    Remove noise due to sensor displacement from the EDA signal.
    
    Parameters
    ----------
    threshold : float, >0
        Threshold to detect the noise
        
    Optional parameters
    -------------------
    
    win_len : float, >0, default = 2
        Length of the window
   
    Returns
    -------
    signal : EvenlySignal
        De-noised signal
            
    """

    def __init__(self, threshold, win_len=2):
        assert threshold > 0, "Threshold value should be positive"
        assert win_len > 0, "Window length value should be positive"
        _Filter.__init__(self, threshold=threshold, win_len=win_len)

    @classmethod
    def algorithm(cls, signal, params):
        threshold = params['threshold']
        win_len = params['win_len']

        # remove fluctiations
        noise = ConvolutionalFilter(irftype='triang', win_len=win_len, normalize=True)(abs(_np.diff(signal)))

        # identify noisy portions
        idx_ok = _np.where(noise <= threshold)[0]

        # fix start and stop of the signal for the following interpolation
        if idx_ok[0] != 0:
            idx_ok = _np.r_[0, idx_ok].astype(int)

        if idx_ok[-1] != len(signal) - 1:
            idx_ok = _np.r_[idx_ok, len(signal) - 1].astype(int)

        denoised = _UnevenlySignal(signal[idx_ok], signal.get_sampling_freq(), x_values=idx_ok, x_type='indices',
                                   duration=signal.get_duration())

        # interpolation
        signal_out = denoised.to_evenly('linear')
        return signal_out


class ConvolutionalFilter(_Filter):
    """
    Filter a signal by convolution with a given impulse response function (IRF).

    Parameters
    ----------
    irftype : str
        Type of IRF to be generated. 'gauss', 'rect', 'triang', 'dgauss', 'custom'.
    win_len : float, >0 (> 8/fsamp for 'gaussian')
        Duration of the generated IRF in seconds (if irftype is not 'custom')
    
    Optional parameters
    -------------------
    irf : numpy.array
        IRF to be used if irftype is 'custom'
    normalize : boolean, default = True
        Whether to normalizes the IRF to have unitary area
    
    Returns
    -------
    signal : EvenlySignal
        Filtered signal

    """

    def __init__(self, irftype, win_len=0, irf=None, normalize=True):
        assert irftype in ['gauss', 'rect', 'triang', 'dgauss', 'custom'],\
            "IRF type must be in ['gauss', 'rect', 'triang', 'dgauss', 'custom']"
        assert irftype == 'custom' or win_len > 0, "Window length value should be positive"
        _Filter.__init__(self, irftype=irftype, win_len=win_len, irf=irf, normalize=normalize)

    # TODO (Andrea): TEST normalization and results
    @classmethod
    def algorithm(cls, signal, params):
        irftype = params["irftype"]
        normalize = params["normalize"]

        fsamp = signal.get_sampling_freq()
        irf = None

        if irftype == 'custom':
            if 'irf' not in params:
                cls.error("'irf' parameter missing.")
                return signal
            else:
                irf = _np.array(params["irf"])
                n = len(irf)
        else:
            if 'win_len' not in params:
                cls.error("'win_len' parameter missing.")
                return signal
            else:
                n = int(params['win_len'] * fsamp)

                if irftype == 'gauss':
                    if n < 8:
                        # TODO (Andrea): test, sometimes it returns nan
                        cls.error(
                            "'win_len' too short to generate a gaussian IRF, expected > " + str(_np.ceil(8 / fsamp)))
                    std = _np.floor(n / 8)
                    irf = _gaussian(n, std)
                elif irftype == 'rect':
                    irf = _np.ones(n)
                elif irftype == 'triang':
                    irf_1 = _np.arange(n // 2)
                    irf_2 = irf_1[-1] - _np.arange(n // 2)
                    if n % 2 == 0:
                        irf = _np.r_[irf_1, irf_2]
                    else:
                        irf = _np.r_[irf_1, irf_1[-1] + 1, irf_2]
                elif irftype == 'dgauss':
                    std = _np.round(n / 8)
                    g = _gaussian(n, std)
                    irf = _np.diff(g)

        # NORMALIZE
        if normalize:
            irf = irf / _np.sum(irf)
            
        signal_ = _np.r_[_np.ones(n) * signal[0], signal, _np.ones(n) * signal[-1]]  # TESTME

        signal_f = _np.convolve(signal_, irf, mode='same')

        signal_out = signal.clone_properties(signal_f[n:-n])
        return signal_out

    @classmethod
    def plot(cls):
        pass


class DeConvolutionalFilter(_Filter):
    """
    Filter a signal by deconvolution with a given impulse response function (IRF).

    Parameters
    ----------
    irf : numpy.array
        IRF used to deconvolve the signal
    
    Optional parameters
    -------------------
    
    normalize : boolean, default = True
        Whether to normalize the IRF to have unitary area
    deconv_method : str, default = 'sps'
        Available methods: 'fft', 'sps'. 'fft' uses the fourier transform, 'sps' uses the scipy.signal.deconvolve
         function
        
    Returns
    -------
    signal : EvenlySignal
        Filtered signal

    """

    def __init__(self, irf, normalize=True, deconv_method='sps'):
        # TODO (Andrea): "check that irf[0]>0 to avoid scipy BUG" is it normal? Need to put a check?
        assert deconv_method in ['fft', 'sps'], "Deconvolution method not valid"
        _Filter.__init__(self, irf=irf, normalize=normalize, deconv_method=deconv_method)

    @classmethod
    def algorithm(cls, signal, params):
        irf = params["irf"]
        normalize = params["normalize"]
        deconvolution_method = params["deconv_method"]

        if normalize:
            irf = irf / _np.sum(irf)
        if deconvolution_method == 'fft':
            l = len(signal)
            fft_signal = _np.fft.fft(signal, n=l)
            fft_irf = _np.fft.fft(irf, n=l)
            out = _np.fft.ifft(fft_signal / fft_irf)
        elif deconvolution_method == 'sps':
            cls.warn('sps based deconvolution needs to be tested. Use carefully.')
            out, _ = _deconvolve(signal, irf)
        else:
            cls.error('Deconvolution method not implemented. Returning original signal.')
            out = signal.get_values()

        out_signal = signal.clone_properties(abs(out))

        return out_signal

    def plot(self):
        _plot(self._params['irf'])



# coding=utf-8
from setuptools import setup

setup(
    name='pyphysio',
    packages=['pyphysio',
              'pyphysio.estimators',
              'pyphysio.filters',
              'pyphysio.indicators',
              'pyphysio.segmentation',
              'pyphysio.tools',
              'pyphysio.tests',
              'pyphysio.sqi',
              ],
    package_data={'pyphysio.tests': ['data/*']},
    version='2.3',
    description='Python library for physiological signals analysis (IBI & HRV, ECG, BVP, EDA, RESP...)',
    author='MPBA FBK',
    author_email='andrea.bizzego@unitn.it',
    url='https://github.com/MPBA/pyphysio',
    keywords=['eda', 'gsr', 'ecg', 'bvp', 'signal', 'analysis', 'physiological', 'pyhrv', 'hrv'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Healthcare Industry',
        'Intended Audience :: Information Technology',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.5',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
    ],
    install_requires=[
        'numpy',
        'scipy',
        'matplotlib',
        'pycwt'
    ],
    requires=[
        'pytest',
    ],
)

print("")
print("")
print("")
print("----------------------------------")
print("|                                |")
print("|  Thanks for using 'pyphysio'!  |")
print("|                                |")
print("----------------------------------")
print("")
print("Remember to cite pyphysio in your publications:")
print("Bizzego et al. (2019) 'pyphysio: A physiological signal processing library for data science approaches in physiology', SoftwareX")
print("")
print("----------------------------------")



# coding=utf-8
from __future__ import division

from . import ph, TestData, np, approx


def test_indicators_eda():
    # %%
    # EDA
    FSAMP = 2048

    eda = ph.EvenlySignal(TestData.eda(), sampling_freq=FSAMP, signal_nature='EDA', start_time=0)

    # pre-processing
    # downsampling
    eda = eda.resample(8)

    # filter
    eda = ph.IIRFilter(fp=0.8, fs=1.1)(eda)

    # estimate driver
    driver = ph.DriverEstim(t1=0.75, t2=2)(eda)
    phasic, tonic, driver_no_peak = ph.PhasicEstim(delta=0.1)(driver)

    # %%
    assert ph.Mean()(phasic) == approx(.0568, abs=0.00005)

    mx = ph.Max()(phasic)
    pks_max = ph.PeaksMax(delta=0.1)(phasic)

    assert (pks_max == mx)

    idx_mx, idx_mn, mx, mn = ph.PeakDetection(delta=0.1)(phasic)
    st, sp = ph.PeakSelection(indices=idx_mx, win_pre=2, win_post=2)(phasic)

    n_peaks = ph.PeaksNum(delta=0.1, win_pre=2, win_post=2)(phasic)

    assert (n_peaks == 24)
    assert (np.sum(st) == 12501)
    assert (np.sum(sp) == 13031)

    # %%
    # FAKE PHASIC
    data = np.zeros(80)
    phasic = ph.EvenlySignal(data, sampling_freq=8, signal_nature='PHA', start_time=0)

    phasic[36] = -0.05
    phasic[37] = 0.12
    phasic[39] = 0.10
    phasic[40] = 0.2
    phasic[41] = 0.10
    phasic[43] = 0.12
    phasic[44] = -0.05

    pks_max = ph.PeaksMax(delta=0.1)(phasic)
    assert (pks_max == 0.2)

    pks_min = ph.PeaksMin(delta=0.1)(phasic)
    assert (pks_min == 0.12)

    pks_mean = ph.PeaksMean(delta=0.1)(phasic)
    assert (pks_mean == np.mean([0.12, 0.2, 0.12]))

    dur_max = ph.DurationMax(delta=0.1, win_pre=2, win_post=2)(phasic)
    assert (dur_max == 0.5)

    dur_min = ph.DurationMin(delta=0.1, win_pre=2, win_post=2)(phasic)
    assert (dur_min == 0.25)

    dur_mean = ph.DurationMean(delta=0.1, win_pre=2, win_post=2)(phasic)
    assert (dur_mean == np.mean([0.25, 0.25, 0.5]))



# coding=utf-8
from __future__ import division

from . import ph
import unittest

__author__ = 'aleb'


def test_annotate_ecg():
    ecg = ph.EvenlySignal(ph.TestData.ecg(), 2048)
    ibi = ph.BeatFromECG()(ecg)

    print(ph.Annotate()(ecg.segment_time(0, 20), []))


if __name__ == '__main__':
    unittest.main()



# coding=utf-8
from __future__ import print_function
from __future__ import division

from . import ph
import numpy as np

import unittest

__author__ = 'aleb'


# noinspection PyArgumentEqualDefault
class GeneralTest(unittest.TestCase):
    def test_issue24(self):
        FSAMP = 100
        n = np.arange(1000)
        t = n / FSAMP
        freq = 1

        # create reference signal
        sinusoid = ph.EvenlySignal(np.sin(2 * np.pi * freq * t), sampling_freq=FSAMP, signal_nature='', start_time=0)

        selection = [10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24]

        # remove some samples
        sinusoid_unevenly = ph.UnevenlySignal(
            np.delete(sinusoid.get_values(), selection),
            sampling_freq=FSAMP, signal_nature='', start_time=None,
            x_values=np.delete(sinusoid.get_times(), selection),
            x_type='instants')

        unev_t = sinusoid_unevenly.get_times()
        even_t = np.delete(sinusoid.get_times(), selection)
        error_sum = np.sum(sinusoid_unevenly.get_times() - np.delete(sinusoid.get_times(), selection))
        error_sum1 = np.sum(unev_t - even_t)
        where_error = np.where(sinusoid_unevenly.get_times() != np.delete(sinusoid.get_times(), selection))
        where_in_even_t = even_t[where_error]
        where_in_unev_t = unev_t[where_error]
        errors = where_in_unev_t - where_in_even_t

        # 69 errors of -0.01

        self.assertEquals(error_sum, error_sum1)
        self.assertEquals(error_sum, 0)
        self.assertEquals(error_sum, np.sum(errors))

    def test_issue22(self):
        values = np.arange(1, 11)
        instan = np.arange(1, 11)

        signal_unevenly = ph.UnevenlySignal(values=values,
                                            sampling_freq=100,
                                            signal_nature='',
                                            start_time=None,
                                            x_values=instan,
                                            x_type='instants')

        one = signal_unevenly.get_times()[0]  # = 1 <=OK
        signal_evenly = signal_unevenly.to_evenly()
        zero = signal_evenly.get_times()[0]  # = 0 <= KO

        self.assertEqual(one, 1)
        self.assertEqual(zero, 1)

        signal_unevenly = ph.UnevenlySignal(values=values,
                                            sampling_freq=100,
                                            signal_nature='',
                                            start_time=0,
                                            x_values=instan,
                                            x_type='instants')

        one = signal_unevenly.get_times()[0]  # = 1 <=OK
        signal_evenly = signal_unevenly.to_evenly()
        zero = signal_evenly.get_times()[0]  # = 0 <= KO

        self.assertEqual(one, 1)
        self.assertEqual(zero, 1)

    def test_issue18_unev(self):
        values = np.arange(1, 11)
        instan = np.arange(1, 11)
        s = ph.UnevenlySignal(values=values,
                              sampling_freq=100,
                              signal_nature='',
                              start_time=None,
                              x_values=instan,
                              x_type='instants')

        pickled = s.pickleable

        s2 = ph.Signal.from_pickleable(pickled)

        self.assertEqual(s.get_sampling_freq(), s2.get_sampling_freq())
        self.assertEqual(s.get_start_time(), s2.get_start_time())
        self.assertEqual(s.get_end_time(), s2.get_end_time())
        self.assertEqual(str(s), str(s2))

    def test_issue18_even(self):
        s = ph.EvenlySignal(values=np.sin(2 * np.pi * np.arange(1000, step=0.01)),
                            sampling_freq=100,
                            signal_nature='',
                            start_time=0)

        pickled = s.pickleable

        s2 = ph.Signal.from_pickleable(pickled)

        self.assertEqual(s.get_sampling_freq(), s2.get_sampling_freq())
        self.assertEqual(s.get_start_time(), s2.get_start_time())
        self.assertEqual(s.get_end_time(), s2.get_end_time())
        self.assertEqual(str(s), str(s2))

    @staticmethod
    def test_issue39():
        ecg = ph.EvenlySignal(ph.TestData.ecg(), sampling_freq=1024)
        algos = [
            ph.PNNx(name="PNN50", threshold=50),
            ph.PNNx(name="PNN20", threshold=20),
            ph.PNNx(name="PNN10", threshold=10),
            ph.Mean(name="Mean"),
            ph.StDev(name="Median"),
        ]

        # %%

        # create fake label
        label = np.zeros(len(ecg))
        label[int(len(ecg) / 2):] = 1
        label[int(3 * len(ecg) / 4):] = 2
        label = ph.EvenlySignal(label, sampling_freq=1024)

        # label based windowing
        label_based = ph.LabelSegments(labels=label)

        # %%

        fixed_length = ph.FixedSegments(step=5, width=20, labels=label)
        result, col_names = ph.fmap(fixed_length, algos, ecg)

        assert result is not None
        assert col_names is not None

        # %%

        # (optional) IIR filtering : remove high frequency noise
        ecg = ph.IIRFilter(fp=45, fs=50, ftype='ellip')(ecg)

        # normalization : normalize data
        ecg = ph.Normalize(norm_method='standard')(ecg)

        # resampling : increase the sampling frequency by cubic interpolation
        ecg = ecg.resample(fout=4096, kind='cubic')
        
        ibi = ph.BeatFromECG()(ecg)

        result, col_names = ph.fmap(label_based(ibi), algos)

        assert result is not None
        assert col_names is not None

    @staticmethod
    def test_issue40():
        # import data and creating a signal

        ecg_data = ph.TestData.ecg()
        fsamp = 2048
        ecg = ph.EvenlySignal(values=ecg_data, sampling_freq=fsamp, signal_nature='ecg')

        # ** Step 1: Filtering and preprocessing **

        # (optional) IIR filtering : remove high frequency noise
        ecg = ph.IIRFilter(fp=45, fs=50, ftype='ellip')(ecg)

        # normalization : normalize data
        ecg = ph.Normalize(norm_method='standard')(ecg)

        # resampling : increase the sampling frequency by cubic interpolation
        ecg = ecg.resample(fout=4096, kind='cubic')

        # ** Step 2: Information Extraction **
        ibi = ph.BeatFromECG()(ecg)

        # define a list of indicators we want to compute
        hrv_indicators = [ph.Mean(name='RRmean'), ph.StDev(name='RRstd'), ph.RMSSD(name='rmsSD'),
                          ph.PowerInBand(name='HF', interp_freq=4, freq_max=0.4, freq_min=0.15, method='ar'),
                          ph.PowerInBand(name='LF', interp_freq=4, freq_max=0.15, freq_min=0.04, method='ar')
                          ]

        # create fake label
        label = np.zeros(1200)
        label[300:600] = 1
        label[900:1200] = 2
        label = ph.EvenlySignal(label, sampling_freq=10, signal_nature='label')

        # fixed length windowing
        fixed_length = ph.FixedSegments(step=5, width=20, labels=label, drop_mixed=False, drop_shorter=True)

        indicators, col_names = ph.fmap(fixed_length(ibi), hrv_indicators)

        print(indicators[:, 0:3])

        # label based windowing
        label_based = ph.LabelSegments(labels=label)

        indicators, col_names = ph.fmap(label_based(ibi), hrv_indicators)

        print(indicators[:, 0:3])

        # custom based windowing
        custom_based = ph.CustomSegments(begins=[0, 30, 60, 90], ends=[30, 60, 90, label.get_duration()], labels=label)

        indicators, col_names = ph.fmap(custom_based, hrv_indicators, ibi)

        print(indicators[:, 0:3])

    @staticmethod
    def test_issue40_10():
        ecg_data = ph.TestData.ecg()
        fsamp = 2048
        ecg = ph.EvenlySignal(values=ecg_data, sampling_freq=fsamp, signal_nature='ecg')

        # resampling : increase the sampling frequency by cubic interpolation
        ecg = ecg.resample(fout=4096, kind='cubic')

        ibi = ph.BeatFromECG()(ecg)

        # create fake label
        label = np.zeros(1200)
        label[300:600] = 1
        label[900:1200] = 2

        label = ph.EvenlySignal(label, sampling_freq=10)

        # label based windowing
        label_based = ph.LabelSegments(labels=label)

        assert len([x for x in label_based(ibi)]) == 4

    @staticmethod
    def test_issue48():
        ecg = ph.EvenlySignal(values=ph.TestData.ecg(), sampling_freq=2048, signal_nature='ecg')
        ibi = ph.BeatFromECG()(ecg)

        # create fake label
        label = np.zeros(1200)
        label[300:600] = 1
        label[900:1200] = 2
        label = ph.EvenlySignal(label, sampling_freq=10, signal_nature='label')

        t_start = [0.5, 15, 88.7]
        t_stop = [5, 21, 110.4]
        custom_segments = ph.CustomSegments(begins=t_start, ends=t_stop, labels=label)

        indicators, col_names = ph.fmap(custom_segments, [ph.Mean()], ibi)

        assert len(indicators) == 2

        assert indicators[0][0] == t_start[0]
        assert indicators[1][0] == t_start[1]

        assert indicators[0][1] == t_stop[0]
        assert indicators[1][1] == t_stop[1]

        assert indicators[0][2] == 0
        assert indicators[1][2] == 0



# coding=utf-8
from __future__ import division
from . import ph, np, TestData, approx


def test_indicators():
    # %%
    FSAMP = 2048
    TSTART = 0

    # %%
    # TEST IBI EXTRACTION FROM ECG
    ecg = ph.EvenlySignal(TestData.ecg(), sampling_freq=FSAMP, signal_nature='ECG', start_time=TSTART)

    ecg = ecg.resample(fout=4096, kind='cubic')

    ibi = ph.BeatFromECG()(ecg)
    # %%
    # TEST Time domain
    assert ph.Mean()(ibi) == approx(.86195, abs=.00005)
    assert ph.StDev()(ibi) == approx(.06027, abs=.00005)
    assert ph.Median()(ibi) == approx(.87133, abs=.00005)
    assert ph.Range()(ibi) == approx(.25488, abs=.00005)
    assert ph.RMSSD()(ibi) == approx(.03260, abs=.00005)
    assert ph.SDSD()(ibi) == approx(.03260, abs=.00005)

    # TEST Frequency domain
    assert ph.PowerInBand(method='welch', interp_freq=4, freq_max=0.04, freq_min=0.00001)(ibi) == approx(127.1227,
                                                                                                         abs=.00005)
    assert ph.PowerInBand(method='welch', interp_freq=4, freq_max=0.15, freq_min=0.04)(ibi) == approx(259.99196,
                                                                                                      abs=.00005)
    assert ph.PowerInBand(method='welch', interp_freq=4, freq_max=0.4, freq_min=0.15)(ibi) == approx(120.18714,
                                                                                                     abs=.00005)

    assert ph.PNNx(threshold=10)(ibi) == approx(.3453, abs=.00005)
    assert ph.PNNx(threshold=25)(ibi) == approx(.2158, abs=.00005)
    assert ph.PNNx(threshold=50)(ibi) == approx(.04316, abs=.00005)

    # %%
    # Test with FAKE IBI
    idx_ibi = np.arange(0, 101, 10).astype(float)
    ibi = ph.UnevenlySignal(np.diff(idx_ibi), x_values=idx_ibi[1:], sampling_freq=10, signal_nature='IBI',
                            x_type='indices')

    assert int(ph.Mean()(ibi)) == 10
    assert int(ph.StDev()(ibi)) == 0

    assert int(ph.Median()(ibi)) == 10
    assert int(ph.Range()(ibi)) == 0

    assert int(ph.RMSSD()(ibi)) == 0
    assert int(ph.SDSD()(ibi)) == 0

    assert ph.AUC()(ibi) == 10

    # TEST Non linear
    assert ph.PoincareSD1()(ibi) == 0
    assert ph.PoincareSD2()(ibi) == 0
    assert ph.ApproxEntropy()(ibi) == 0
    assert ph.SampleEntropy()(ibi) == 0

    ibi[-1] = 10.011

    assert ph.Mean()(ibi) != 10
    assert ph.StDev()(ibi) != 0

    assert int(ph.Median()(ibi)) == 10

    assert ph.Range()(ibi) != 0

    assert ph.RMSSD()(ibi) != 0
    assert ph.SDSD()(ibi) != 0

    assert ph.PNNx(threshold=10)(ibi) != 0
    assert ph.PNNx(threshold=25)(ibi) == 0
    assert ph.PNNx(threshold=50)(ibi) == 0

    assert ph.PoincareSD1SD2()(ibi) != 1



# coding=utf-8
from __future__ import division

from ..BaseIndicator import Indicator as _Indicator
from ..tools.Tools import PSD as PSD
import numpy as _np

__author__ = 'AleB'


class InBand(_Indicator):
    """
    Extract the PSD of a given frequency band
    

    Parameters
    ----------
    freq_min : float, >0
        Left bound of the frequency band
    freq_max : float, >0
        Right bound of the frequency band
    method : 'ar', 'welch' or 'fft'
        Method to estimate the PSD
        
    Additional parameters
    ---------------------
    For the PSD (see pyphysio.tools.Tools.PSD), for instance:
        
    interp_freq : float, >0
        Frequency used to (re-)interpolate the signal

    Returns
    -------
    freq : numpy array
        Frequencies in the frequency band
    psd : float
        Power Spectrum Density in the frequency band
    """

    def __init__(self, freq_min, freq_max, method, **kwargs):
        _Indicator.__init__(self, freq_min=freq_min, freq_max=freq_max, method=method, **kwargs)

    @classmethod
    def algorithm(cls, data, params):
        freq, spec = PSD(**params)(data)
        # freq is sorted so
        i_min = _np.searchsorted(freq, params["freq_min"])
        i_max = _np.searchsorted(freq, params["freq_max"])
        return freq[i_min:i_max], spec[i_min:i_max]


class PowerInBand(_Indicator):
    """
    Estimate the power in given frequency band

    Parameters
    ----------
    freq_min : float, >0
        Left bound of the frequency band
    freq_max : float, >0
        Right bound of the frequency band
    method : 'ar', 'welch' or 'fft'
        Method to estimate the PSD
        
    Additional parameters
    ---------------------
    For the PSD (see pyphysio.tools.Tools.PSD):
        
    interp_freq : float, >0
        Frequency used to (re-)interpolate the signal

    Returns
    -------
    power : float
        Power in the frequency band
    """

    def __init__(self, freq_min, freq_max, method, **kwargs):
        _Indicator.__init__(self, freq_min=freq_min, freq_max=freq_max, method=method, **kwargs)

    @classmethod
    def algorithm(cls, data, params):
        freq, powers = InBand(**params)(data)
        return _np.sum(powers)


class PeakInBand(_Indicator):
    """
    Estimate the peak frequency in a given frequency band

    Parameters
    ----------
    freq_min : float, >0
        Left bound of the frequency band
    freq_max : float, >0
        Right bound of the frequency band
    method : 'ar', 'welch' or 'fft'
        Method to estimate the PSD
        
    Additional parameters
    ---------------------
    For the PSD (see pyphysio.tools.Tools.PSD):
        
    interp_freq : float, >0
        Frequency used to (re-)interpolate the signal

    Returns
    -------
    peak : float
        Peak frequency
    """

    def __init__(self, freq_min, freq_max, method, **kwargs):
        _Indicator.__init__(self, freq_min=freq_min, freq_max=freq_max, method=method, **kwargs)
    
    @classmethod
    def algorithm(cls, data, params):
        freq, power = InBand(**params)(data)
        return freq[_np.argmax(power)]

